"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
from re import error as RegexError

import ply.lex as lex

from inmanta.ast import LocatableString, Range
from inmanta.ast.constraint.expression import Regex
from inmanta.ast.variables import Reference
from inmanta.parser import ParserException

states = (("mls", "exclusive"),)

keyworldlist = [
    "typedef",
    "as",
    "matching",
    "entity",
    "extends",
    "end",
    "in",
    "implementation",
    "for",
    "index",
    "implement",
    "using",
    "when",
    "and",
    "or",
    "not",
    "true",
    "false",
    "import",
    "is",
    "defined",
    "dict",
    "null",
    "undef",
    "parents",
    "if",
    "else",
]
literals = [":", "[", "]", "(", ")", "=", ",", ".", "{", "}", "?", "*"]
reserved = {k: k.upper() for k in keyworldlist}

# List of token names.   This is always required
tokens = ["INT", "FLOAT", "ID", "CID", "SEP", "STRING", "MLS", "MLS_END", "CMP_OP", "REGEX", "REL", "PEQ"] + sorted(
    list(reserved.values())
)


def t_ID(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[a-zA-Z_][a-zA-Z_0-9-]*"
    t.type = reserved.get(t.value, "ID")  # Check for reserved words
    if t.value[0].isupper():
        t.type = "CID"
    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    (s, e) = lexer.lexmatch.span()
    start = end - (e - s)

    t.value = LocatableString(
        t.value, Range(lexer.inmfile, lexer.lineno, start, lexer.lineno, end), lexer.lexpos, lexer.namespace
    )
    return t


def t_SEP(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[:]{2}"
    return t


def t_REL(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"--|->|<-"
    return t


def t_CMP_OP(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"!=|==|>=|<=|<|>"
    return t


def t_PEQ(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[+]="
    return t


def t_COMMENT(t: lex.LexToken) -> None:  # noqa: N802
    r"\#.*?\n"
    t.lexer.lineno += 1
    t.lexer.linestart = t.lexer.lexpos
    pass


def t_JCOMMENT(t: lex.LexToken) -> None:  # noqa: N802
    r"\//.*?\n"
    t.lexer.lineno += 1
    t.lexer.linestart = t.lexer.lexpos
    pass


def t_begin_mls(t: lex.LexToken) -> lex.LexToken:
    r'["]{3}'
    t.lexer.begin("mls")
    t.type = "MLS"

    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    (s, e) = lexer.lexmatch.span()
    start = end - (e - s)

    t.value = LocatableString("", Range(lexer.inmfile, lexer.lineno, start, lexer.lineno, end), lexer.lexpos, lexer.namespace)

    return t


def t_mls_end(t: lex.LexToken) -> lex.LexToken:
    r'.*["]{3}'
    t.lexer.begin("INITIAL")
    t.type = "MLS_END"
    value = t.value[:-3]

    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    (s, e) = lexer.lexmatch.span()
    start = end - (e - s)

    t.value = LocatableString(
        value, Range(lexer.inmfile, lexer.lineno, start, lexer.lineno, end), lexer.lexpos, lexer.namespace
    )

    return t


def t_mls_MLS(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r".+"
    return t


def t_FLOAT(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[-]?[0-9]*[.][0-9]+"
    t.value = float(t.value)
    return t


def t_INT(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"[-]?[0-9]+"
    t.value = int(t.value)
    return t


def t_STRING(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"(\"([^\\\"]|\\.)*\")|(\'([^\\\']|\\.)*\')"
    t.value = bytes(t.value[1:-1], "utf-8").decode("unicode_escape")
    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    (s, e) = lexer.lexmatch.span()
    start = end - (e - s)

    t.value = LocatableString(
        t.value, Range(lexer.inmfile, lexer.lineno, start, lexer.lineno, end), lexer.lexpos, lexer.namespace
    )

    return t


def t_REGEX(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"/([^/\\]|\\.)+/"
    value = Reference("self")  # anonymous value
    try:
        expr = Regex(value, t.value[1:-1])
        t.value = expr
        return t
    except RegexError as error:
        end = t.lexer.lexpos - t.lexer.linestart + 1
        (s, e) = t.lexer.lexmatch.span()
        start = end - (e - s)

        r: Range = Range(t.lexer.inmfile, t.lexer.lineno, start, t.lexer.lineno, end)
        raise ParserException(r, t.value, "Regex error in %s: '%s'" % (t.value, error))


# Define a rule so we can track line numbers
def t_newline(t: lex.LexToken) -> None:  # noqa: N802
    r"\n+"
    t.lexer.lineno += len(t.value)
    t.lexer.linestart = t.lexer.lexpos


def t_mls_newline(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    r"\n+"
    t.lexer.lineno += len(t.value)
    t.lexer.linestart = t.lexer.lexpos
    t.type = "MLS"
    return t


# A string containing ignored characters (spaces and tabs)
t_ignore = " \t"
t_mls_ignore = ""

# Error handling rule


def t_ANY_error(t: lex.LexToken) -> lex.LexToken:  # noqa: N802
    lexer = t.lexer

    end = lexer.lexpos - lexer.linestart + 1
    char: str = t.value[0]
    r: Range = Range(lexer.inmfile, lexer.lineno, end, lexer.lineno, end + 1)
    raise ParserException(r, char, "Illegal character '%s'" % char)


# Build the lexer
lexer = lex.lex()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""


class AnyType(object):
    """
        Supertype for objects that are an instance of all types
    """

    pass


class Unknown(AnyType):
    """
        An instance of this class is used to indicate that this value can not be determined yet.

        :param source: The source object that can determine the value
    """

    def __init__(self, source):
        self.source = source

    def __iter__(self):
        return iter([])


class NoneValue(object):
    def __eq__(self, other):
        return isinstance(other, NoneValue)

    def __hash__(self):
        return hash(None)

    def __str__(self):
        return "null"

    def __repr__(self):
        return "null"



from inmanta.plugins import plugin


@plugin
def test_list(instance: "test_1774::Test") -> "list":
    return [instance]


@plugin
def test_dict(instance: "test_1774::Test") -> "dict":
    return {"instance": instance}



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: bart@inmanta.com
"""

import logging
from subprocess import CalledProcessError

import pytest

from inmanta import env
from utils import LogSequence


def test_basic_install(tmpdir):
    """If this test fails, try running "pip uninstall lorem dummy-yummy iplib" before running it.
    """

    env_dir1 = tmpdir.mkdir("env1").strpath

    with pytest.raises(ImportError):
        import lorem  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    venv1._install(["lorem"])
    import lorem  # NOQA

    lorem.sentence()

    with pytest.raises(ImportError):
        import yummy  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    venv1.install_from_list(["dummy-yummy"])
    import yummy  # NOQA

    with pytest.raises(ImportError):
        import iplib  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    try:
        venv1.install_from_list(
            ["lorem == 0.1.1", "dummy-yummy", "iplib@git+https://github.com/bartv/python3-iplib", "lorem", "iplib >=0.0.1"]
        )
    except CalledProcessError as ep:
        print(ep.stdout)
        raise
    import iplib  # NOQA


def test_install_fails(tmpdir, caplog):
    venv = env.VirtualEnv(tmpdir)
    venv.use_virtual_env()
    caplog.clear()
    package_name = "non-existing-pkg-inmanta"
    exception_raised = False
    try:
        venv.install_from_list([package_name])
    except Exception:
        exception_raised = True

    assert exception_raised
    log_sequence = LogSequence(caplog)
    log_sequence.contains("inmanta.env", logging.ERROR, f"requirements: {package_name}")


def test_install_package_already_installed_in_parent_env(tmpdir):
    venv = env.VirtualEnv(tmpdir)
    venv.use_virtual_env()
    installed_packages = [
        p for p in env.VirtualEnv._get_installed_packages(venv._parent_python).keys() if p != "pip" and p != "setuptools"
    ]
    random_package = installed_packages[0]
    venv.install_from_list([random_package])

    # Assert not installed in virtual_python venv
    assert random_package not in env.VirtualEnv._get_installed_packages(venv.virtual_python).keys()


def test_req_parser(tmpdir):
    url = "git+https://github.com/bartv/python3-iplib"
    at_url = "iplib@" + url
    egg_url = url + "#egg=iplib"

    e = env.VirtualEnv(tmpdir)
    name, u = e._parse_line(url)
    assert name is None
    assert u == url

    name, u = e._parse_line(at_url)
    assert name == "iplib"
    assert u == egg_url

    e._parse_line(egg_url)
    assert name == "iplib"
    assert u == egg_url


def test_gen_req_file(tmpdir):
    e = env.VirtualEnv(tmpdir)
    req = ["lorem == 0.1.1", "lorem > 0.1", "dummy-yummy", "iplib@git+https://github.com/bartv/python3-iplib", "lorem"]

    req_lines = [x for x in e._gen_requirements_file(req).split("\n") if len(x) > 0]
    assert len(req_lines) == 3


def test_remove_requirements_present_in_parent_env(tmpdir):
    v = env.VirtualEnv(tmpdir)
    v.use_virtual_env()

    # Verify parsing works
    reqs = [
        "lorem == 0.1.1",
        "lorem > 0.1",
        "lorem >= 1.1,<1.5",
        "dummy-yummy",
        "iplib@git+https://github.com/bartv/python3-iplib",
    ]
    assert v._remove_requirements_present_in_parent_env(reqs) == reqs

    installed_packages = {
        k: v for k, v in env.VirtualEnv._get_installed_packages(v._parent_python).items() if k != "pip" and k != "setuptools"
    }
    package = next(iter(installed_packages.keys()))
    package_version = installed_packages[package]
    splitted_package_version = package_version.split(".", maxsplit=1)
    newer_package_version = f"{int(splitted_package_version[0]) + 1}.{splitted_package_version[1]}"

    # Package is present in parent venv and constraint is met
    reqs = [f"{package}=={package_version}", "non_existing_package==1.1.1"]
    reqs_after_removal = ["non_existing_package==1.1.1"]
    assert v._remove_requirements_present_in_parent_env(reqs) == reqs_after_removal

    reqs = [f"{package}>={package_version},<{newer_package_version}", "non_existing_package==1.1.1"]
    assert v._remove_requirements_present_in_parent_env(reqs) == reqs_after_removal

    # Package is present in parent venv, but constraint isn't met
    reqs = [f"{package}=={newer_package_version}", "non_existing_package==1.1.1"]
    assert v._remove_requirements_present_in_parent_env(reqs) == reqs

    reqs = [f"{package}>{package_version}", "non_existing_package==1.1.1"]
    assert v._remove_requirements_present_in_parent_env(reqs) == reqs



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
from typing import Optional

from inmanta.reporter import InfluxReporter
from inmanta.server import SLICE_METRICS
from inmanta.server import config as opt
from inmanta.server import protocol

LOGGER = logging.getLogger(__name__)


class MetricsService(protocol.ServerSlice):
    """Slice managing metrics collector"""

    def __init__(self) -> None:
        super(MetricsService, self).__init__(SLICE_METRICS)
        self._influx_db_reporter: Optional[InfluxReporter] = None

    async def start(self) -> None:
        self.start_metric_reporters()
        await super().start()

    async def stop(self) -> None:
        await super().stop()
        self.stop_metric_reporters()

    def stop_metric_reporters(self) -> None:
        if self._influx_db_reporter:
            self._influx_db_reporter.stop()
            self._influx_db_reporter = None

    def start_metric_reporters(self) -> None:
        if opt.influxdb_host.get():
            self._influx_db_reporter = InfluxReporter(
                server=opt.influxdb_host.get(),
                port=opt.influxdb_port.get(),
                database=opt.influxdb_name.get(),
                username=opt.influxdb_username.get(),
                password=opt.influxdb_password.get(),
                reporting_interval=opt.influxdb_interval.get(),
                autocreate_database=True,
                tags=opt.influxdb_tags.get(),
            )
            self._influx_db_reporter.start()



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
import re
import typing
import urllib
from typing import Dict, Optional

from inmanta.agent.cache import AgentCache

from . import local, remote

if typing.TYPE_CHECKING:
    from inmanta.agent.io.local import IOBase


LOGGER = logging.getLogger(__name__)


def parse_agent_uri(uri: str) -> typing.Tuple[str, Dict[str, Optional[str]]]:
    """
        Parse an agent uri and return the settings

        :attr uri: The uri to parse
        :return: (scheme, config)
    """
    parts = urllib.parse.urlparse(uri)
    config: Dict[str, Optional[str]] = {}
    scheme = "local"

    if parts.query != "":
        items = urllib.parse.parse_qs(parts.query)
        for key, values in items.items():
            config[key] = values[0]

    if parts.scheme != "":
        scheme = parts.scheme

    if parts.netloc != "":
        match = re.search(r"^(?:(?P<user>[^\s@]+)@)?(?P<host>[^\s:]+)(?::(?P<port>[\d]+))?", parts.netloc)
        if match is None:
            raise ValueError()
        config.update(match.groupdict())

    if parts.path != "" and parts.scheme == "" and parts.netloc == "":
        if parts.path == "localhost":
            scheme = "local"
        else:
            scheme = "ssh"
            config.update({"host": parts.path, "port": None, "user": None})

    return scheme, config


def _get_io_class(scheme: str) -> typing.Type[local.IOBase]:
    """
        Get an IO instance.
    """
    if scheme == "local":
        return local.LocalIO

    elif scheme == "ssh":
        return remote.SshIO

    raise Exception("%s scheme is not supported" % scheme)


def _get_io_instance(uri: str) -> "IOBase":
    scheme, config = parse_agent_uri(uri)
    io_class = _get_io_class(scheme)
    LOGGER.debug("Using io class %s for uri %s (%s, %s)", io_class, uri, scheme, config)
    io = io_class(uri, config)
    return io


def get_io(cache: AgentCache, uri: str, version: int) -> "IOBase":
    """
        Get an IO instance for the given uri and version
    """
    if cache is None:
        io = _get_io_instance(uri)
    else:
        io = cache.get_or_else(uri, lambda version: _get_io_instance(uri), call_on_delete=lambda x: x.close(), version=version)
    return io



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import itertools
import logging
import os
import time
from collections import deque
from typing import TYPE_CHECKING, Any, Deque, Dict, Iterator, List, Sequence, Set, Tuple

from inmanta import plugins
from inmanta.ast import Anchor, CompilerException, CycleExcpetion, Location, MultiException, RuntimeException
from inmanta.ast.entity import Entity
from inmanta.ast.statements import DefinitionStatement, TypeDefinitionStatement
from inmanta.ast.statements.define import (
    DefineEntity,
    DefineImplement,
    DefineIndex,
    DefineRelation,
    DefineTypeConstraint,
    DefineTypeDefault,
)
from inmanta.ast.type import TYPES, Type
from inmanta.const import LOG_LEVEL_TRACE
from inmanta.execute.proxy import UnsetException
from inmanta.execute.runtime import (
    DelayedResultVariable,
    ExecutionContext,
    ExecutionUnit,
    Instance,
    QueueScheduler,
    Resolver,
    Waiter,
)
from inmanta.execute.tracking import ModuleTracker

if TYPE_CHECKING:
    from inmanta.compiler import Compiler  # noqa: F401
    from inmanta.ast import Statement, BasicBlock, NamedType  # noqa: F401


DEBUG = True
LOGGER = logging.getLogger(__name__)


MAX_ITERATIONS = 10000


class Scheduler(object):
    """
        This class schedules statements for execution
    """

    def __init__(self, track_dataflow: bool = False):
        self.track_dataflow: bool = track_dataflow
        self.types: Dict[str, Type] = {}

    def freeze_all(self, exns: List[CompilerException]) -> None:
        for t in [t for t in self.types.values() if isinstance(t, Entity)]:
            t.final(exns)

        instances: List[Instance] = self.types["std::Entity"].get_all_instances()

        for i in instances:
            i.final(exns)

    def dump(self, type: str = "std::Entity") -> None:
        instances: List[Instance] = self.types[type].get_all_instances()

        for i in instances:
            i.dump()

    def verify_done(self) -> List[Instance]:
        instances: List[Instance] = self.types["std::Entity"].get_all_instances()
        notdone = []
        for i in instances:
            if not i.verify_done():
                notdone.append(i)

        return notdone

    def get_types(self) -> Dict[str, Type]:
        return self.types

    def dump_not_done(self) -> None:
        for i in self.verify_done():
            i.dump()

    def sort_entities(self, entity_map: Dict[str, DefineEntity]) -> List[DefineEntity]:
        out: List[DefineEntity] = []
        loopstack: Set[str] = set()
        while len(entity_map) > 0:
            workon = next(iter(entity_map.keys()))
            self.do_sort_entities(entity_map, workon, out, loopstack)
        return out

    def do_sort_entities(
        self, entity_map: Dict[str, DefineEntity], name: str, acc: List[DefineEntity], loopstack: Set[str]
    ) -> None:
        nexte = entity_map[name]
        try:
            del entity_map[name]
            loopstack.add(name)
            for p in nexte.get_full_parent_names():
                if p in loopstack:
                    raise CycleExcpetion(nexte, p)
                if p in entity_map:
                    self.do_sort_entities(entity_map, p, acc, loopstack)
            loopstack.remove(name)
            acc.append(nexte)
        except CycleExcpetion as ce:
            ce.add(nexte)
            raise

    def define_types(self, compiler: "Compiler", statements: Sequence["Statement"], blocks: Sequence["BasicBlock"]) -> None:
        """
            This is the first compiler stage that defines all types_and_impl
        """
        # get all relevant stmts
        definitions: List["DefinitionStatement"] = [d for d in statements if isinstance(d, DefinitionStatement)]
        others: List["Statement"] = [d for d in statements if not isinstance(d, DefinitionStatement)]

        if not len(others) == 0:
            raise Exception("others not empty %s" % repr(others))

        # collect all  types and impls
        types_and_impl = {}

        # set primitive types
        compiler.get_ns().set_primitives(TYPES)

        # all stmts contributing types and impls
        newtypes: List[Tuple[str, NamedType]] = [
            k for k in [t.register_types() for t in definitions if isinstance(t, TypeDefinitionStatement)] if k is not None
        ]

        for (name, type_symbol) in newtypes:
            types_and_impl[name] = type_symbol

        # now that we have objects for all types, populate them
        implements = [t for t in definitions if isinstance(t, DefineImplement)]
        other_definitions = [t for t in definitions if not isinstance(t, DefineImplement)]
        entities: Dict[str, DefineEntity] = {t.fullName: t for t in other_definitions if isinstance(t, DefineEntity)}
        type_constraints = [t for t in other_definitions if isinstance(t, DefineTypeConstraint)]
        typedefaults = [t for t in other_definitions if isinstance(t, DefineTypeDefault)]
        other_definitions = [
            t for t in other_definitions if not isinstance(t, (DefineEntity, DefineTypeDefault, DefineTypeConstraint))
        ]
        indices = [t for t in other_definitions if isinstance(t, DefineIndex)]
        other_definitions = [t for t in other_definitions if not isinstance(t, DefineIndex)]

        # first type constraints so attribute defaults can be type checked
        for tc in type_constraints:
            tc.evaluate()

        # then entities, so we have inheritance
        # parents first
        for entity in self.sort_entities(entities):
            entity.evaluate()

        for typedefault in typedefaults:
            typedefault.evaluate()

        for other in other_definitions:
            other.evaluate()

        # indices late, as they require all attributes
        for index in indices:
            index.evaluate()

        # lastly the implements, as they require implementations
        for implement in implements:
            implement.evaluate()

        compiler.plugins = {k: v for k, v in types_and_impl.items() if isinstance(v, plugins.Plugin)}
        types = {k: v for k, v in types_and_impl.items() if isinstance(v, Type)}

        # normalize plugins
        for p in compiler.plugins.values():
            p.normalize()

        # give type info to all types, to normalize blocks inside them
        for t in types.values():
            t.normalize()

        # normalize root blocks
        for block in blocks:
            block.normalize()

        self.types = {k: v for k, v in types_and_impl.items() if isinstance(v, Type)}

    def anchormap(
        self, compiler: "Compiler", statements: Sequence["Statement"], blocks: Sequence["BasicBlock"]
    ) -> Sequence[Tuple[Location, Location]]:
        prev = time.time()

        # first evaluate all definitions, this should be done in one iteration
        self.define_types(compiler, statements, blocks)

        # relations are also in blocks
        not_relation_statements: Iterator[Statement] = (s for s in statements if not isinstance(s, DefineRelation))

        anchors: Iterator[Anchor] = (
            anchor
            for container in itertools.chain(not_relation_statements, blocks)  # container: Union[Statement, BasicBlock]
            for anchor in container.get_anchors()
            if anchor is not None
        )

        rangetorange = [(anchor.get_location(), anchor.resolve()) for anchor in anchors]
        rangetorange = [(f, t) for f, t in rangetorange if t is not None]

        now = time.time()
        LOGGER.debug("Anchormap took %f seconds", now - prev)

        return rangetorange

    def run(self, compiler: "Compiler", statements: Sequence["Statement"], blocks: Sequence["BasicBlock"]) -> bool:
        """
            Evaluate the current graph
        """
        prev = time.time()

        # first evaluate all definitions, this should be done in one iteration
        self.define_types(compiler, statements, blocks)

        # give all loose blocks an empty XC
        # register the XC's as scopes
        # All named scopes are now present

        for block in blocks:
            res = Resolver(block.namespace, self.track_dataflow)
            xc = ExecutionContext(block, res)
            block.context = xc
            block.namespace.scope = xc
            block.warn_shadowed_variables()

        # setup queues
        # queue for runnable items
        basequeue: Deque[Waiter] = deque()
        # queue for RV's that are delayed
        waitqueue: Deque[DelayedResultVariable[Any]] = deque()
        # queue for RV's that are delayed and had no effective waiters when they were first in the waitqueue
        zerowaiters: Deque[DelayedResultVariable[Any]] = deque()
        # queue containing everything, to find hanging statements
        all_statements: Set[Waiter] = set()

        # Wrap in object to pass around
        queue = QueueScheduler(compiler, basequeue, waitqueue, self.types, all_statements)

        # emit all top level statements
        for block in blocks:
            block.context.emit(queue.for_tracker(ModuleTracker(block)))

        # start an evaluation loop
        i = 0
        count = 0
        max_iterations = int(os.getenv("INMANTA_MAX_ITERATIONS", MAX_ITERATIONS))
        while i < max_iterations:
            now = time.time()

            # check if we can stop the execution
            if len(basequeue) == 0 and len(waitqueue) == 0 and len(zerowaiters) == 0:
                break
            else:
                i += 1

            LOGGER.debug(
                "Iteration %d (e: %d, w: %d, p: %d, done: %d, time: %f)",
                i,
                len(basequeue),
                len(waitqueue),
                len(zerowaiters),
                count,
                now - prev,
            )
            prev = now

            # evaluate all that is ready
            while len(basequeue) > 0:
                next = basequeue.popleft()
                try:
                    next.execute()
                    all_statements.discard(next)
                    count = count + 1
                except UnsetException as e:
                    # some statements don't know all their dependencies up front,...
                    next.waitfor(e.get_result_variable())

            # all safe stmts are done
            progress = False

            # find a RV that has waiters, so freezing creates progress
            while len(waitqueue) > 0 and not progress:
                next_rv = waitqueue.popleft()
                if next_rv.hasValue:
                    # already froze itself
                    continue
                if next_rv.get_progress_potential() <= 0:
                    zerowaiters.append(next_rv)
                elif next_rv.get_waiting_providers() > 0:
                    # definitely not done
                    # drop from queue
                    # will requeue when value is added
                    next_rv.unqueue()
                else:
                    # freeze it and go to next iteration, new statements will be on the basequeue
                    LOGGER.log(LOG_LEVEL_TRACE, "Freezing %s", next_rv)
                    next_rv.freeze()
                    progress = True

            # no waiters in waitqueue,...
            # see if any zerowaiters have become gotten waiters
            if not progress:
                zerowaiters_tmp = [w for w in zerowaiters if not w.hasValue]
                waitqueue = deque(w for w in zerowaiters_tmp if w.get_progress_potential() > 0)
                queue.waitqueue = waitqueue
                zerowaiters = deque(w for w in zerowaiters_tmp if w.get_progress_potential() <= 0)
                while len(waitqueue) > 0 and not progress:
                    LOGGER.debug("Moved zerowaiters to waiters")
                    next_rv = waitqueue.popleft()
                    if next_rv.get_waiting_providers() > 0:
                        next_rv.unqueue()
                    else:
                        LOGGER.log(LOG_LEVEL_TRACE, "Freezing %s", next_rv)
                        next_rv.freeze()
                        progress = True

            # no one waiting anymore, all done, freeze and finish
            if not progress:
                LOGGER.debug("Finishing statements with no waiters")
                while len(zerowaiters) > 0:
                    next_rv = zerowaiters.pop()
                    next_rv.freeze()

        now = time.time()
        LOGGER.info(
            "Iteration %d (e: %d, w: %d, p: %d, done: %d, time: %f)",
            i,
            len(basequeue),
            len(waitqueue),
            len(zerowaiters),
            count,
            now - prev,
        )

        if i == max_iterations:
            raise CompilerException(f"Could not complete model, max_iterations {max_iterations} reached.")

        # now = time.time()
        # print(now - prev)
        # end evaluation loop
        # self.dump_not_done()
        # print(basequeue, waitqueue)
        # dumpHangs()
        # self.dump()
        # rint(len(self.types["std::Entity"].get_all_instances()))

        excns: List[CompilerException] = []
        self.freeze_all(excns)

        if len(excns) == 0:
            pass
        elif len(excns) == 1:
            raise excns[0]
        else:
            raise MultiException(excns)

        if all_statements:
            stmt = None
            for st in all_statements:
                if isinstance(st, ExecutionUnit):
                    stmt = st
                    break

            assert stmt is not None

            raise RuntimeException(stmt.expression, "not all statements executed %s" % all_statements)
        # self.dump("std::File")

        return True



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import inspect
import logging
import pkgutil
import types
import uuid
from asyncio import Semaphore
from typing import Optional

import asyncpg
import pytest
from asyncpg import PostgresSyntaxError

import inmanta.db.versions
from data.db import versions
from data.db_with_invalid_versions import invalid_versions
from inmanta import data
from inmanta.data import CORE_SCHEMA_NAME, schema
from inmanta.data.schema import InvalidSchemaVersion, TableNotFound, Version, create_schemamanager
from utils import log_contains


async def run_updates_and_verify(
    get_columns_in_db_table, schema_manager: schema.DBSchema, current: Optional[int] = None, prefix: str = ""
):
    async def update_function1(connection):
        await connection.execute(f"CREATE TABLE public.{prefix}tab(id integer primary key, val varchar NOT NULL);")

    async def update_function2(connection):
        await connection.execute(f"ALTER TABLE public.{prefix}tab DROP COLUMN val;")

    current_db_version = await schema_manager.get_current_version()
    assert current is None or current_db_version == current
    if current_db_version is None:
        current_db_version = 0
    version_update1 = current_db_version + 1
    version_update2 = current_db_version + 2
    update_function_map = [Version(f"v{version_update1}", update_function1), Version(f"v{version_update2}", update_function2)]

    await schema_manager._update_db_schema(update_function_map)

    assert (await schema_manager.get_current_version()) == version_update2
    assert sorted(["id"]) == sorted(await get_columns_in_db_table("tab"))


async def get_core_version(postgresql_client):
    dbm = schema.DBSchema(CORE_SCHEMA_NAME, inmanta.db.versions, postgresql_client)
    try:
        return await dbm.get_current_version()
    except TableNotFound:
        return 0


async def assert_core_untouched(postgresql_client, corev=0):
    """
    Verify abscence of side-effect leaks to other cases
    """
    dbm = schema.DBSchema(CORE_SCHEMA_NAME, inmanta.db.versions, postgresql_client)
    current_db_version = await dbm.get_current_version()
    assert current_db_version == corev


@pytest.mark.asyncio
async def test_dbschema_clean(postgresql_client: asyncpg.Connection, get_columns_in_db_table, hard_clean_db, caplog):
    with caplog.at_level(logging.INFO):
        dbm = schema.DBSchema("test_dbschema_clean", inmanta.db.versions, postgresql_client)
        with pytest.raises(TableNotFound):
            await dbm.get_current_version()
        await dbm.ensure_self_update()
        await run_updates_and_verify(get_columns_in_db_table, dbm, 0)
        await assert_core_untouched(postgresql_client)

        log_contains(caplog, "inmanta.data.schema.schema:test_dbschema_clean", logging.INFO, "Creating schema version table")


@pytest.mark.asyncio
async def test_dbschema_unclean(postgresql_client: asyncpg.Connection, get_columns_in_db_table, hard_clean_db):
    dbm = schema.DBSchema("test_dbschema_unclean", inmanta.db.versions, postgresql_client)
    current_db_version = await dbm.ensure_self_update()
    assert current_db_version == 0

    await dbm.set_current_version(5)
    current_db_version = await dbm.get_current_version()
    assert current_db_version == 5

    await run_updates_and_verify(get_columns_in_db_table, dbm, current_db_version)
    await assert_core_untouched(postgresql_client)


@pytest.mark.asyncio
async def test_dbschema_update_legacy_1(
    postgresql_client: asyncpg.Connection, get_columns_in_db_table, hard_clean_db, hard_clean_db_post
):
    await postgresql_client.execute(
        """
-- Table: public.schemaversion
CREATE TABLE IF NOT EXISTS public.schemaversion(
    id uuid PRIMARY KEY,
    current_version integer NOT NULL UNIQUE
);
"""
    )
    dbm = schema.DBSchema("test_l1", inmanta.db.versions, postgresql_client)
    current_db_version = await dbm.get_legacy_version()
    assert current_db_version == 0
    await dbm.ensure_self_update()
    await run_updates_and_verify(get_columns_in_db_table, dbm, current_db_version)


@pytest.mark.asyncio
async def test_dbschema_update_legacy_2(
    postgresql_client: asyncpg.Connection, get_columns_in_db_table, hard_clean_db, hard_clean_db_post
):
    await postgresql_client.execute(
        """
-- Table: public.schemaversion
CREATE TABLE IF NOT EXISTS public.schemaversion(
    id uuid PRIMARY KEY,
    current_version integer NOT NULL UNIQUE
);
"""
    )
    await postgresql_client.execute("INSERT INTO public.schemaversion(id, current_version) VALUES ($1, $2);", uuid.uuid4(), 1)

    dbm = schema.DBSchema(CORE_SCHEMA_NAME, inmanta.db.versions, postgresql_client)
    current_db_version = await dbm.get_legacy_version()
    assert current_db_version == 1
    await dbm.ensure_self_update()
    await run_updates_and_verify(get_columns_in_db_table, dbm, current_db_version)


@pytest.mark.asyncio
async def test_dbschema_update_db_schema(postgresql_client, get_columns_in_db_table, hard_clean_db, hard_clean_db_post):

    db_schema = schema.DBSchema("test_dbschema_update_db_schema", inmanta.db.versions, postgresql_client)
    await db_schema.ensure_self_update()

    await run_updates_and_verify(get_columns_in_db_table, db_schema, 0)

    db_schema = schema.DBSchema("test_dbschema_update_db_schema_1", inmanta.db.versions, postgresql_client)

    await run_updates_and_verify(get_columns_in_db_table, db_schema, 0, prefix="C")


@pytest.mark.asyncio
async def test_dbschema_update_db_schema_failure(postgresql_client, get_columns_in_db_table):
    corev = await get_core_version(postgresql_client)
    db_schema = schema.DBSchema("test_dbschema_update_db_schema_failure", inmanta.db.versions, postgresql_client)
    await db_schema.ensure_self_update()

    async def update_function(connection):
        # Syntax error should trigger database rollback
        await connection.execute("CREATE TABE public.tab(id integer primary key, val varchar NOT NULL);")

    current_db_version = await db_schema.get_current_version()
    new_db_version = current_db_version + 1
    update_function_map = [Version(f"v{new_db_version}", update_function)]

    with pytest.raises(PostgresSyntaxError):
        await db_schema._update_db_schema(update_function_map)

    # Assert rollback
    assert (await db_schema.get_current_version()) == current_db_version
    assert (
        await postgresql_client.fetchval(
            "SELECT table_name FROM information_schema.tables " "WHERE table_schema='public' AND table_name='tab'"
        )
    ) is None

    async def update_function2(connection):
        # Fix syntax issue
        await connection.execute("CREATE TABLE public.tab(id integer primary key, val varchar NOT NULL);")

    update_function_map = [Version(f"v{new_db_version}", update_function2)]
    await db_schema._update_db_schema(update_function_map)

    # Assert update
    assert (await db_schema.get_current_version()) == new_db_version
    assert sorted(["id", "val"]) == sorted(await get_columns_in_db_table("tab"))
    await assert_core_untouched(postgresql_client, corev)


def make_version(nr, fct):
    return Version(f"v{nr}", fct)


def make_versions(idx, *fcts):
    return [make_version(idx + i, fct) for i, fct in enumerate(fcts)]


@pytest.mark.asyncio
async def test_dbschema_partial_update_db_schema_failure(postgresql_client, get_columns_in_db_table):
    corev = await get_core_version(postgresql_client)
    db_schema = schema.DBSchema("test_dbschema_partial_update_db_schema_failure", inmanta.db.versions, postgresql_client)
    await db_schema.ensure_self_update()

    async def update_function_good(connection):
        # Fix syntax issue
        await connection.execute("CREATE TABLE public.taba(id integer primary key, val varchar NOT NULL);")

    async def update_function_bad(connection):
        # Syntax error should trigger database rollback
        await connection.execute("CREATE TABLE public.tabb(id integer primary key, val varchar NOT NULL);")
        await connection.execute("CREATE TABE public.tab(id integer primary key, val varchar NOT NULL);")

    async def update_function_good2(connection):
        # Fix syntax issue
        await connection.execute("CREATE TABLE public.tabc(id integer primary key, val varchar NOT NULL);")

    current_db_version = await db_schema.get_current_version()
    if current_db_version is None:
        current_db_version = 0

    update_function_map = make_versions(
        current_db_version + 1, update_function_good, update_function_bad, update_function_good2
    )

    with pytest.raises(PostgresSyntaxError):
        await db_schema._update_db_schema(update_function_map)

    # Assert rollback
    assert (await db_schema.get_current_version()) == current_db_version + 1
    assert (
        await postgresql_client.fetchval(
            "SELECT table_name FROM information_schema.tables " "WHERE table_schema='public' AND table_name='taba'"
        )
    ) is not None

    assert (
        await postgresql_client.fetchval(
            "SELECT table_name FROM information_schema.tables " "WHERE table_schema='public' AND table_name='tabb'"
        )
    ) is None

    assert (
        await postgresql_client.fetchval(
            "SELECT table_name FROM information_schema.tables " "WHERE table_schema='public' AND table_name='tabc'"
        )
    ) is None

    await assert_core_untouched(postgresql_client, corev)


@pytest.mark.asyncio
async def test_dbschema_get_dct_with_update_functions():
    module_names = [modname for _, modname, ispkg in pkgutil.iter_modules(data.PACKAGE_WITH_UPDATE_FILES.__path__) if not ispkg]
    for module_name in module_names:
        module = __import__(data.PACKAGE_WITH_UPDATE_FILES.__name__ + "." + module_name, fromlist=["update"])
        if module.DISABLED:
            module_names.remove(module_name)
    all_versions = [int(mod_name[1:]) for mod_name in module_names]

    db_schema = schema.DBSchema(CORE_SCHEMA_NAME, data.PACKAGE_WITH_UPDATE_FILES, None)
    update_function_map = await db_schema._get_update_functions()
    assert sorted(all_versions) == [v.version for v in update_function_map]
    for version in update_function_map:
        assert version.version >= 0
        assert isinstance(version.function, types.FunctionType)
        assert version.function.__name__ == "update"
        assert inspect.getfullargspec(version.function)[0] == ["connection"]


@pytest.mark.asyncio
async def test_multi_upgrade_lockout(postgresql_pool, get_columns_in_db_table, hard_clean_db):
    async with postgresql_pool.acquire() as postgresql_client:
        async with postgresql_pool.acquire() as postgresql_client2:

            # schedule 3 updates, hang on second, unblock one, verify, unblock other, verify
            corev = await get_core_version(postgresql_client)

            db_schema = schema.DBSchema("test_multi_upgrade_lockout", inmanta.db.versions, postgresql_client)
            db_schema2 = schema.DBSchema("test_multi_upgrade_lockout", inmanta.db.versions, postgresql_client2)
            await db_schema.ensure_self_update()

            lock = Semaphore(0)

            async def update_function_a(connection):
                # Fix syntax issue
                await connection.execute("CREATE TABLE public.taba(id integer primary key, val varchar NOT NULL);")

            async def update_function_b(connection):
                # Syntax error should trigger database rollback
                await lock.acquire()
                await connection.execute("CREATE TABLE public.tabb(id integer primary key, val varchar NOT NULL);")

            async def update_function_c(connection):
                # Fix syntax issue
                await connection.execute("CREATE TABLE public.tabc(id integer primary key, val varchar NOT NULL);")

            current_db_version = await db_schema.get_current_version()
            if current_db_version is None:
                current_db_version = 0

            update_function_map = make_versions(current_db_version + 1, update_function_a, update_function_b, update_function_c)

            r1 = asyncio.ensure_future(db_schema._update_db_schema(update_function_map))
            r2 = asyncio.ensure_future(db_schema2._update_db_schema(update_function_map))

            both = asyncio.as_completed([r1, r2]).__iter__()

            await asyncio.sleep(0.1)
            first = next(both)
            lock.release()
            await first

            # second one doesn't even hit the lock, as it never sees schema version 0
            # lock.release()
            second = next(both)
            await second

            # Assert done
            assert (await db_schema.get_current_version()) == current_db_version + 3
            assert (
                await postgresql_client.fetchval(
                    "SELECT table_name FROM information_schema.tables " "WHERE table_schema='public' AND table_name='taba'"
                )
            ) is not None

            assert (
                await postgresql_client.fetchval(
                    "SELECT table_name FROM information_schema.tables " "WHERE table_schema='public' AND table_name='tabb'"
                )
            ) is not None

            assert (
                await postgresql_client.fetchval(
                    "SELECT table_name FROM information_schema.tables " "WHERE table_schema='public' AND table_name='tabc'"
                )
            ) is not None

            await assert_core_untouched(postgresql_client, corev)


@pytest.mark.asyncio
async def test_dbschema_get_dct_filter_disabled():
    db_schema = schema.DBSchema(CORE_SCHEMA_NAME, versions, None)
    update_function_map = await db_schema._get_update_functions()
    assert 2 not in [v.version for v in update_function_map]
    for version in update_function_map:
        assert version.version >= 0
        assert isinstance(version.function, types.FunctionType)
        assert version.function.__name__ == "update"
        assert inspect.getfullargspec(version.function)[0] == ["connection"]


@pytest.mark.asyncio
async def test_dbschema_update_db_downgrade(postgresql_client):
    schema_name = "test_dbschema_update_db_downgrade"
    SCHEMA_VERSION_TABLE = "schemamanager"
    await postgresql_client.execute(create_schemamanager)

    db_schema = schema.DBSchema(schema_name, inmanta.db.versions, postgresql_client)
    update_function_map = await db_schema._get_update_functions()
    original_version = len(update_function_map) + 1
    await postgresql_client.execute(
        f"INSERT INTO {SCHEMA_VERSION_TABLE} (name, current_version) VALUES ($1, $2)", schema_name, original_version
    )
    with pytest.raises(InvalidSchemaVersion):
        await db_schema.ensure_db_schema()
    current_db_version = await db_schema.get_current_version()
    assert original_version == current_db_version


@pytest.mark.asyncio
async def test_dbschema_get_dct_filter_invalid_names(caplog):
    db_schema = schema.DBSchema(CORE_SCHEMA_NAME, invalid_versions, None)
    update_function_map = await db_schema._get_update_functions()
    assert len(update_function_map) == 0
    log_contains(caplog, "inmanta.data.schema", logging.WARNING, "V2 doesn't match the expected pattern")
    log_contains(caplog, "inmanta.data.schema", logging.WARNING, "ver1 doesn't match the expected pattern")
    log_contains(caplog, "inmanta.data.schema", logging.WARNING, "ver1a doesn't match the expected pattern")
    log_contains(caplog, "inmanta.data.schema", logging.WARNING, "version3 doesn't match the expected pattern")
    log_contains(caplog, "inmanta.data.schema", logging.WARNING, "v1b doesn't match the expected pattern")



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import pkgutil
import re
from types import ModuleType
from typing import Any, Callable, Coroutine, List, Optional, Tuple

from asyncpg import Connection, UndefinedTableError
from asyncpg.transaction import Transaction

# Name of core schema in the DB schema verions
CORE_SCHEMA_NAME = "core"

LOGGER = logging.getLogger(__name__)

LEGACY_SCHEMA_VERSION_TABLE = "schemaversion"
SCHEMA_VERSION_TABLE = "schemamanager"

create_schemamanager = """
-- Table: public.schemamanager
CREATE TABLE IF NOT EXISTS public.schemamanager (
    name varchar PRIMARY KEY,
    current_version integer NOT NULL
);
"""


class TableNotFound(Exception):
    """ Raised when a table is not found in the database """

    pass


class InvalidSchemaVersion(Exception):
    """ Raised when an invalid database version is found """

    pass


class Version(object):
    """ Internal representation of a version """

    def __init__(self, name: str, function: Callable[[Connection], Coroutine[Any, Any, None]]):
        self.name = name
        self.function = function
        self.version = self.parse(name)

    @classmethod
    def parse(cls, name: str) -> int:
        return int(name[1:])


class DBSchema(object):
    """
    Schema Manager, ensures the schema is up to date.

    Concurrent updates are safe
    """

    def __init__(self, name: str, package: ModuleType, connection: Connection) -> None:
        """

        :param name: unique name for this schema, best equal to extension name and used as prefix for all table names
        :param package: a python package, containing modules with name v%(version)s.py.
            Each module contains a method `async def update(connection: asyncpg.connection) -> None:`
        :param connection: asyncpg connection
        """
        self.name = name
        self.package = package
        self.connection = connection
        self.logger = LOGGER.getChild(f"schema:{self.name}")

    async def ensure_db_schema(self) -> None:
        current_version_db_schema = await self.ensure_self_update()
        update_functions = await self._get_update_functions()
        desired_version = update_functions[-1].version
        if desired_version > current_version_db_schema:
            await self._update_db_schema(update_functions)
        elif desired_version < current_version_db_schema:
            raise InvalidSchemaVersion(
                f"Desired database version {desired_version} is lower "
                f"than the current version {current_version_db_schema}, downgrading is not supported"
            )

    async def ensure_self_update(self) -> int:
        try:
            await self.get_legacy_version()
            await self._legacy_migration()
        except TableNotFound:
            # No legacy table, proceed
            pass

        try:
            return await self.get_current_version()
        except TableNotFound:
            self.logger.info("Creating schema version table")
            # create table
            await self.connection.execute(create_schemamanager)
            return 0

    async def _legacy_migration(self) -> None:
        """
        Migration to new schema management:
        1- as long as the legacy schemaversion table exists, no other operation is allowed by DBSchema
        2- takes a lock on the legacy schemaversion table to ensure exclusivity
        3- migrates the existing version to the new table, for the core slice
        4- drops legacy table
        """
        self.logger.info("Migrating from old schema management to new schema management")
        # tx begin
        async with self.connection.transaction():
            # lock legacy table => if gone -> continue
            try:
                await self.connection.execute(f"LOCK TABLE {LEGACY_SCHEMA_VERSION_TABLE} IN ACCESS EXCLUSIVE MODE")
            except UndefinedTableError:
                self.logger.info("Second process is preforming a database update as well.")
                return
            # get_legacy_version, under lock
            legacy_version_db_schema = await self.get_legacy_version()

            if legacy_version_db_schema > 0:
                self.logger.info("Creating schema version table and setting core version to %d", legacy_version_db_schema)
                # create table
                await self.connection.execute(create_schemamanager)
                await self.connection.execute(
                    f"INSERT INTO {SCHEMA_VERSION_TABLE}(name, current_version) VALUES ($1, $2)",
                    CORE_SCHEMA_NAME,
                    legacy_version_db_schema,
                )
            else:
                self.logger.info("Creating schema version table")
                # create table
                await self.connection.execute(create_schemamanager)

            await self.connection.execute(f"DROP TABLE {LEGACY_SCHEMA_VERSION_TABLE}")

    async def _update_db_schema(self, update_functions: List[Version]) -> None:
        """
        Main update function

        Wrapped in outer transaction, that holds a lock on the schemamanager table.
        Each version update is wrapped in a subtransaction.

        When a subtransaction fails, it is rolled back.
        The outer transaction is committed at that point.

        This logic requires manual transaction management,
        as the exception is propagated over the transaction boundary without causing rollback.
        """
        # outer transaction
        outer: Optional[Transaction]
        outer = self.connection.transaction()
        try:
            # enter transaction
            await outer.start()
            # get lock
            await self.connection.execute(f"LOCK TABLE {SCHEMA_VERSION_TABLE} IN ACCESS EXCLUSIVE MODE")
            # get current version again, in transaction this time
            try:
                sure_db_schema = await self.get_current_version()
            except TableNotFound:
                self.logger.exception("Schemamanager table disappeared, should not occur.")
                raise
            # get relevant updates
            updates = [v for v in update_functions if v.version > sure_db_schema]
            for version in updates:
                try:
                    # wrap in subtransaction
                    async with self.connection.transaction():
                        # actual update sequence
                        self.logger.info("Updating database schema to version %d", version.version)
                        update_function = version.function
                        await update_function(self.connection)
                        # also set version, outer tx will always contain consistent version
                        await self.set_current_version(version.version)
                    # commit subtx
                except Exception:
                    # update failed, subtransaction already rolled back
                    self.logger.exception("Database schema update to version %d failed", version.version)
                    # commit outer
                    await outer.commit()
                    # unset it, to prevent double commit
                    outer = None
                    # propagate excn
                    raise
        except Exception:
            # an exception, from either outer transaction (before subtransaction) or subtransaction
            if outer is not None:
                # subtransaction did not set None, so abort
                await outer.rollback()
                outer = None
            raise
        finally:
            # if the tx is still there, all is good
            if outer is not None:
                await outer.commit()

    async def get_legacy_version(self) -> int:
        try:
            version = await self.connection.fetchrow(f"select current_version from {LEGACY_SCHEMA_VERSION_TABLE}")
        except UndefinedTableError as e:
            raise TableNotFound() from e
        if version is None:
            return 0
        return version["current_version"]

    async def get_current_version(self) -> int:
        try:
            version = await self.connection.fetchrow(
                f"select current_version from {SCHEMA_VERSION_TABLE} where name=$1", self.name
            )
        except UndefinedTableError as e:
            raise TableNotFound() from e
        if version is None:
            return 0
        return version["current_version"]

    async def set_current_version(self, version: int) -> None:
        await self.connection.execute(
            f"INSERT INTO {SCHEMA_VERSION_TABLE}(name, current_version) "
            "VALUES ($1, $2) ON CONFLICT(name) DO UPDATE SET current_version=$2",
            self.name,
            version,
        )

    async def _get_update_functions(self) -> List[Version]:
        module_names = [modname for _, modname, ispkg in pkgutil.iter_modules(self.package.__path__) if not ispkg]

        def get_modules(mod_name: str) -> Tuple[str, ModuleType]:
            fq_module_name = self.package.__name__ + "." + mod_name
            return mod_name, __import__(fq_module_name, fromlist=["update"])

        def make_version(mod_name: str, module: ModuleType) -> Version:
            update_function = module.update
            return Version(mod_name, update_function)

        pattern = re.compile("^v[0-9]+$")

        filtered_module_names = []
        for module_name in module_names:
            if not pattern.match(module_name):
                LOGGER.warning(
                    f"Database schema version file name {module_name} "
                    f"doesn't match the expected pattern: v<version_number>.py, skipping it"
                )
            else:
                filtered_module_names.append(module_name)

        modules_with_names = [get_modules(mod_name) for mod_name in filtered_module_names]
        filtered_modules = [(module_name, module) for module_name, module in modules_with_names if not module.DISABLED]

        version = [make_version(name, mod) for name, mod in filtered_modules]

        return sorted(version, key=lambda x: x.version)



"""
    Copyright 2020 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from typing import Union

import pytest

from inmanta.ast import Namespace
from inmanta.ast.entity import Entity
from inmanta.execute.dataflow import (
    AssignableNode,
    AttributeNode,
    AttributeNodeReference,
    InstanceNode,
    Node,
    NodeReference,
    ValueNode,
)


def entity_instance(entity: str) -> InstanceNode:
    node: InstanceNode = InstanceNode([])
    node.entity = Entity(entity, Namespace("__config__", Namespace("__root_ns__")))
    return node


@pytest.mark.parametrize(
    "instance,expected_repr",
    [
        (ValueNode(42), "42"),
        (ValueNode("42"), "'42'"),
        (ValueNode(42).reference(), "42"),
        (ValueNode("Hello World!"), repr("Hello World!")),
        (AssignableNode("x"), "x"),
        (AssignableNode("x").reference(), "x"),
        (AttributeNodeReference(AttributeNodeReference(AssignableNode("x").reference(), "y"), "z"), "x.y.z"),
        (entity_instance("MyEntity"), "__config__::MyEntity instance"),
        (entity_instance("MyEntity").reference(), "__config__::MyEntity instance"),
        (AttributeNode(entity_instance("MyEntity"), "n"), "attribute n on __config__::MyEntity instance"),
    ],
)
def test_dataflow_repr(instance: Union[Node, NodeReference], expected_repr: str) -> None:
    assert repr(instance) == expected_repr
    assert str(instance) == expected_repr
